using System;
using System.ComponentModel;
using System.Collections;
using System.Diagnostics;
using System.Drawing;
using System.Windows.Forms;

namespace BrainfuckStudio
{
	/// <summary>
	/// Zusammenfassung fr MemTable.
	/// </summary>
	public class MemTable : System.Windows.Forms.Panel, IObserver
	{
		private Font font;				// used font
		private int numOfLines;			// number of visible lines
		private MemoryBand Memory;		// reference to memory

		private VScrollBar scrollbar = new VScrollBar();

		private int firstVisibleLine;	// index of first line to show
		private int lastVisibleLine;	// index of last line to show

		private static bool bRepaintFromMem = false;	// helper for scrolling
		

		/**
		 * standard constructor
		 **/
		public MemTable()
		{
			this.firstVisibleLine = 0;

			// create the scrollbar
			scrollbar.Parent = this;
			scrollbar.Location = new Point(0, this.Width - System.Windows.Forms.SystemInformation.VerticalScrollBarWidth
				- System.Windows.Forms.SystemInformation.Border3DSize.Width - 5);
			scrollbar.Size = new Size(System.Windows.Forms.SystemInformation.VerticalScrollBarWidth, this.Height);
			scrollbar.Minimum = 0;
			//scrollbar.Maximum = Memory.MemLength- scrollbar.SmallChange;
			scrollbar.SmallChange = 1;
			scrollbar.LargeChange =  5;
			scrollbar.Scroll += new System.Windows.Forms.ScrollEventHandler(OnScrolling);
			scrollbar.BringToFront();

			// create my font
			CreateFont();

			// handle mouse wheel
			this.MouseWheel += new MouseEventHandler(MemTable_MouseWheel);
			this.Click += new EventHandler(MemTable_Click);
		}

		void OnScrolling(Object obj, System.Windows.Forms.ScrollEventArgs sea)
		{
			this.Invalidate();
		}

		private void CreateFont()
		{
			font = new Font("Courier New", 10);
		}

		protected override void OnResize(EventArgs e)
		{
			if (Memory == null) return;

			base.OnResize (e);

			scrollbar.Location = new Point(this.Width - 
				System.Windows.Forms.SystemInformation.VerticalScrollBarWidth -
				System.Windows.Forms.SystemInformation.Border3DSize.Width, 0);
			scrollbar.Size = new Size(SystemInformation.VerticalScrollBarWidth, 
				this.Height - SystemInformation.Border3DSize.Height);

			float f = font.GetHeight();
			numOfLines = (int)(this.Height / font.GetHeight());
			numOfLines++;
			this.lastVisibleLine = this.firstVisibleLine + numOfLines;

			this.Invalidate();
		}


		protected override void OnPaint(System.Windows.Forms.PaintEventArgs e)
		{
			// call base paint
			base.OnPaint (e);

			if (Memory == null) return;			// bye

			int curPos = Memory.CurrentPos;
			int firstLine, lastLine;

			// the repaint signal came from the memory band, so take care if we have to scroll ourselves
			if (bRepaintFromMem)
			{
				if (curPos < this.firstVisibleLine || curPos > this.lastVisibleLine)
				{
					int temp = curPos - this.numOfLines / 2;
					if (temp < 0) temp = 0;
					if (temp > Memory.MemLength) temp = Memory.MemLength - 1;
					scrollbar.Minimum = Math.Max(0, temp - 100);
					scrollbar.Maximum = Math.Min(Memory.MemLength, temp + 100);
					scrollbar.Value = curPos;
					this.firstVisibleLine = temp;
					this.lastVisibleLine = Math.Min(Memory.MemLength-1, temp + this.numOfLines);
				}
				bRepaintFromMem = false;
				firstLine = this.firstVisibleLine; 
				lastLine = this.lastVisibleLine; 
			}
			else 
			{
				firstLine = scrollbar.Value;
				lastLine = Math.Min(Memory.MemLength - 1, firstLine + numOfLines);
			}

			// draw the lines
			try 
			{
				DrawLines(firstLine, lastLine, e);
			}
			catch (IndexOutOfRangeException RangeEx)
			{
				Logger.getLogger().LogWithTime("MemTable :" + RangeEx.Message);
			}
			
			e.Graphics.Dispose();

		}	

		/**
		 * Draw the lines
		 **/
		private void DrawLines(int firstLine, int lastLine, PaintEventArgs e)  
		{
			string s = "";
			string sTemp;

			int curPos = Memory.CurrentPos;
			// draw the first part of lines in black till current position
			for (int i = firstLine; i < curPos; i++)
			{
				sTemp = String.Format("{0:D6}  {1:D3}  0x{2:X2}  ", i,
					Memory.Memory[i],
					Memory.Memory[i]);
				s += sTemp + "\n";
			}
			e.Graphics.DrawString(s, font, System.Drawing.Brushes.Black,0,0);

			// draw the line with the current position
			s = sTemp = "";
			if (Memory.DataType == MemoryBand.DATATYPE.SIGNED_SHORT)
				s = String.Format("{0:D6}  {1:D3}  0x{2:X2}  ", curPos,
					(short)Memory.Memory[curPos],
					(short)Memory.Memory[curPos]);
			else if (Memory.DataType == MemoryBand.DATATYPE.SIGNED_BYTE)
				s = String.Format("{0:D6}  {1:D3}  0x{2:X2}  ", curPos,
					(sbyte)Memory.Memory[curPos],
					(sbyte)Memory.Memory[curPos]);
			else
				s = String.Format("{0:D6}  {1:D3}  0x{2:X2}  ", curPos,
					Memory.Memory[curPos],
					Memory.Memory[curPos]);
			e.Graphics.DrawString(s, font, System.Drawing.Brushes.Yellow,0, (curPos - firstLine) * font.GetHeight());

			// draw the rest of the lines
			s = sTemp = "";
			for (int i = curPos + 1; i <= lastLine; i++)
			{
				sTemp = String.Format("{0:D6}  {1:D3}  0x{2:X2}  ", i,
					Memory.Memory[i],
					Memory.Memory[i]);
				s += sTemp + "\n";
			}
			e.Graphics.DrawString(s, font, System.Drawing.Brushes.Black,0, (curPos - firstLine + 1) * font.GetHeight());
		}

		private void MemTable_MouseWheel(object sender, MouseEventArgs e)
		{
			int numLines = e.Delta * SystemInformation.MouseWheelScrollLines / 120;
			try 
			{
				scrollbar.Value -= numLines;
			}
			catch (System.ArgumentException){ }		// important but no need to handle it in a special way
			
			this.Invalidate();
		}
		
		public MemoryBand Memoryband
		{
			set { Memory = value; }
		}

		private void MemTable_Click(object sender, EventArgs e)
		{
			this.Focus();
		}

		#region IObserver Member

		public void Update(ISubject sender)
		{
			if (this.Memory == null) return;
			bRepaintFromMem = true;
			this.Invalidate();

		}

		#endregion

	}
}
