using System;
using System.Collections;
using System.Windows.Forms;

namespace BrainfuckStudio
{
	/*
	 * Interface definitions for observer pattern
	 */
	public interface ISubject
	{
		void Register(IObserver observer);
		void UnRegister(IObserver observer);
		void Notify();
	}

	public interface IObserver
	{
		void Update(ISubject sender);
	}


	/// <summary>
	/// Zusammenfassung fr MemoryBand.
	/// </summary>
	public class MemoryBand : ISubject
	{
		public const int STATUS_JUMPOUT = 1;
		public const int STATUS_JUMPLOOP = -1;
		public const int STATUS_INPUTREQUIRED = 2;

		public enum DATATYPE {BYTE, SIGNED_BYTE, SHORT, SIGNED_SHORT, SIGNED_INT};

		private DATATYPE currentType;
		private int[] memory;
		private int currentPos;

		// normally a bf array is 30000 cells, but a turing machine is infinite
		// true means the array is dynamically expanded if necessary; otherwise 
		// it is set to standard length of 30000
		//public bool ExpandDynamic = false;		// TO DO!

		// hold all oberserver
		private ArrayList clients = new ArrayList(5);

		// if this is true, no clients will be notified
		public bool bDisableNotify = false;

		// reference to the programm output box
		public TextBox OutputBox;
		// reference to the input box from which ascii values are read in
		public TextBox InputBox;

		// count how many commands where executed
		private int commandcount = 0;

		/// <summary>
		/// Constructor
		/// </summary>
		public MemoryBand()
		{
			memory = new int[30000];
			currentType = DATATYPE.SIGNED_INT;		// standard cell is signed int
		}

		public MemoryBand(TextBox inbox, TextBox outbox) : this()
		{
			OutputBox = outbox;
			InputBox = inbox;
		}


		/// <summary>
		/// Reset the memory to zero
		/// </summary>
		public void Reset()
		{
			for (int i = 0; i < this.memory.Length; i++)
				memory[i] = 0;
			currentPos = 0;
			commandcount = 0;
			Notify();
		}

		public int InterpretCommand(char command)
		{
			commandcount++;
			switch (command)
			{
				case '>':
					// flip around if necessary
					if (++currentPos >= this.MemLength)
						currentPos = 0;
					break;

				case '<':
					// flip around if necessary
					if (--currentPos < 0)
						currentPos = this.MemLength - 1;
					break;

				case '+':
				switch (currentType)
				{
					case DATATYPE.SIGNED_BYTE:
					{
						if (memory[currentPos] == System.SByte.MaxValue)
							memory[currentPos] = System.SByte.MinValue-1;
					} break;
					case DATATYPE.BYTE:
					{
						if (memory[currentPos] == System.Byte.MaxValue)
							memory[currentPos] = System.Byte.MinValue-1;
					} break;
					case DATATYPE.SIGNED_SHORT:
					{
						if (memory[currentPos] == System.Int16.MaxValue)
							memory[currentPos] = System.Int16.MinValue-1;
					} break;
					case DATATYPE.SHORT:
					{
						if (memory[currentPos] == System.UInt16.MaxValue)
							memory[currentPos] = System.UInt16.MinValue-1;
					} break;

					default: break;
				}
					memory[currentPos]++;
					break;

				case '-':
				switch (currentType)
				{
					case DATATYPE.SIGNED_BYTE:
					{
						if (memory[currentPos] == System.SByte.MinValue)
							memory[currentPos] = System.SByte.MaxValue+1;
					} break;
					case DATATYPE.BYTE:
					{
						if (memory[currentPos] == System.Byte.MinValue)
							memory[currentPos] = System.Byte.MaxValue+1;
					} break;
					case DATATYPE.SIGNED_SHORT:
					{
						if (memory[currentPos] == System.Int16.MinValue)
							memory[currentPos] = System.Int16.MaxValue+1;
					} break;
					case DATATYPE.SHORT:
					{
						if (memory[currentPos] == System.UInt16.MinValue)
							memory[currentPos] = System.UInt16.MaxValue+1;
					} break;

					default: break;
				}
					memory[currentPos]--;
					break;

				case '.':
					if (this.OutputBox == null) break;
					int val = memory[currentPos];
					char c = ((char)val).ToString()[0];
					// till now special handling
					if (val == 13)
						this.OutputBox.AppendText("\r\n");
					else
						this.OutputBox.AppendText(c.ToString());
					
					break;

				case ',':
					return STATUS_INPUTREQUIRED;

				case '[':
					if (memory[currentPos] == 0) return STATUS_JUMPOUT;
					break;

				case ']':
					if (memory[currentPos] != 0) return STATUS_JUMPLOOP;
					break;

				default: break;
			}
			Notify();
			return 0;
		}



		#region ISubject Member

		public void Register(IObserver observer)
		{
			clients.Add(observer);
		}

		public void UnRegister(IObserver observer)
		{
			clients.Remove(observer);
		}

		public void Notify()
		{
			if (this.bDisableNotify) return;

			for ( int i = 0; i < clients.Count; i++ )
				((IObserver) clients[i]).Update(this);
		}

		#endregion

		#region Properties
		public int CurrentPos 
		{
			get { return this.currentPos; }
			set { if ( (value >= 0) && (value < memory.Length)) currentPos = value; }
		}

		public int[] Memory
		{
			get { return this.memory; }
		}

		public int MemLength
		{
			get { return memory.Length; }
		}

		public int NumberOfCommands
		{
			get { return commandcount; }
		}

		public DATATYPE DataType
		{
			set { currentType = value; }
			get { return currentType; }
		}
		
		#endregion
	}
}
